/* $Id: action.c,v 1.37 2010/03/05 01:45:26 oohara Exp $ */

#include <stdio.h>

#include "world.h"
#include "creature.h"
#include "dungeon.h"
#include "save.h"

#include "action.h"

static int player_pickup(world *wp);

static int
player_pickup(world *wp)
{
  int i;
  int j;
  char buf[128];
  creature *player = NULL;

  if (wp == NULL)
    return 1;

  if (creature_is_dead(wp, wp->player_id))
    return 0;
  player = wp->cr[wp->player_id];

  for (i = 0; i < wp->obj_size; i++)
  {
    if (object_is_dead(wp, i))
      continue;
    if ((wp->obj[i]->x != player->x)
        || (wp->obj[i]->y != player->y))
      continue;
    for (j = 0; j < 9; j++)
    {
      if ((wp->inv[j][0] >= 0)
          && (wp->inv[j][0] != wp->obj[i]->type))
        continue;
      sprintf(buf, "player picks up object type %d (#%d)",
              wp->obj[i]->type,
              wp->obj[i]->id);
      add_log(wp, buf);
      wp->inv[j][0] = wp->obj[i]->type;
      wp->inv[j][1] += wp->obj[i]->quantity;
      wp->obj[i]->quantity = 0;
      break;
    }
  }

  return 0;
}

/* return 1 if the action is legal, 0 otherwise */
int
action_check(world *wp, int creature_id, int *act)
{
  int x;
  int y;
  creature *who = NULL;

  if (wp == NULL)
  {
    fprintf(stderr, "action_check: wp is NULL\n");
    return 0;
  }
  if (act == NULL)
  {
    fprintf(stderr, "action_check: act is NULL\n");
    return 0;
  }
  if ((creature_id < 0) || (creature_id >= wp->cr_size))
  {
    fprintf(stderr, "action_check: strange creature_id (%d)\n",
            creature_id);
    return 0;
  }

  if (creature_is_dead(wp, creature_id))
    return 0;
  who = wp->cr[creature_id];

  switch(act[0])
  {
  case ACTION_STAY:
    return 1;
    break;
  case ACTION_WALK:
    /* act
     * [1]: dx
     * [2]: dy
     */
    if ((act[1] == 0) && (act[2] == 0))
      return 1;
    if ((act[1] < -1) || (act[1] > 1))
      return 0;
    if ((act[2] < -1) || (act[2] > 1))
      return 0;
    x = who->x + act[1];
    y = who->y + act[2];
    if ((x < 0) || (x >= wp->size_x))
      return 0;
    if ((y < 0) || (y >= wp->size_y))
      return 0;
    if (grid_blocks_creature(wp, x, y))
      return 0;
    return 1;
    break;
  case ACTION_USE_OBJECT:
    if (who->id == wp->player_id)
      return 1;
    return 0;
    break;
  case ACTION_SAVE:
    if (who->id == wp->player_id)
      return 1;
    return 0;
    break;
  default:
    return 0;
    break;
  }

  fprintf(stderr, "action_check: should not reach here\n");
  return 0;
}

/* return the amount of time taken by the action */
int
action_exec(world *wp, int creature_id)
{
  int ut;
  int ut_taken;
  int i;
  int x;
  int y;
  char buf[128];
  creature *who = NULL;

  if (wp == NULL)
  {
    fprintf(stderr, "action_exec: wp is NULL\n");
    return 0;
  }
  if ((creature_id < 0) || (creature_id >= wp->cr_size))
  {
    fprintf(stderr, "action_exec: strange creature_id (%d)\n",
            creature_id);
    return 0;
  }

  if (creature_is_dead(wp, creature_id))
    return 0;
  who = wp->cr[creature_id];

  if (who->act[0] == 0)
    return 0;

  ut = 4;
  ut_taken = 0;
  if (!action_check(wp, who->id, who->act))
  {
    /* trying to do something illegal */
    return ut;
  }

  switch(who->act[0])
  {
  case ACTION_STAY:
    ut_taken = ut;
    break;
  case ACTION_WALK:
    x = who->x + who->act[1];
    y = who->y + who->act[2];
    ut_taken = ut;

    if ((!grid_blocks_creature(wp, x, y))
        && (!grid_has_creature(wp, x, y)))
    {
      if (who->id == wp->player_id)
      {
        wp->camera_x = x;
        wp->camera_y = y;
      }
      if ((wp->cursor_x == who->x)
          && (wp->cursor_y == who->y))
      {
        wp->cursor_x = x;
        wp->cursor_y = y;
      }
      who->x = x;
      who->y = y;
      break;
    }

    for (i = 0; i < wp->cr_size; i++)
    {
      if (i == who->id)
        continue;
      if (creature_is_dead(wp, i))
        continue;
      if ((wp->cr[i]->x == x) && (wp->cr[i]->y == y))
      {
        wp->cr[i]->hp -= 1;
        sprintf(buf, "creature #%d hits creature #%d",
                who->id, i);
        add_log(wp, buf);
        if (creature_is_dead(wp, i))
        {
          sprintf(buf, "creature #%d kills creature #%d",
                who->id, i);
          add_log(wp, buf);
        }
        break;
      }
    }
    break;
  case ACTION_USE_OBJECT:
    ut_taken = ut;
    sprintf(buf, "creature #%d uses inventory #%d",
            who->id, who->act[1]);
    add_log(wp, buf);
    break;
  case ACTION_SAVE:
    ut_taken = 0;
    if (save_game(wp) == 0)
    {
      sprintf(buf, "game saved successfully");
      add_log(wp, buf);
    }
    else
    {
      sprintf(buf, "game save FAILED");
      add_log(wp, buf);
    }
    break;
  default:
    /* trying to do something undefined */
    ut_taken = ut;
    break;
  }

  if (who->id == wp->player_id)
  {
    player_pickup(wp);
  }

  for (i = 0; i < 4; i++)
    who->act[i] = 0;

  return ut_taken;
}
