/* $Id: dungeon.c,v 1.37 2010/03/05 08:23:11 oohara Exp $ */

#include <stdio.h>
/* realloc, abs */
#include <stdlib.h>

#include "world.h"
#include "creature.h"
#include "list-int.h"
#include "util.h"

#include "dungeon.h"

int
distance(int ax, int ay, int bx, int by)
{
  int dx;
  int dy;

  dx = bx - ax;
  dy = by - ay;

  if (abs(dx) > abs(dy))
    return abs(dx);
  
  return abs(dy);
}

int
compare_wait(list_int *p, list_int *q, void *data)
{
  creature *who_p;
  creature *who_q;
  world *wp = (world *) data;

  if (wp == NULL)
    return 0;

  who_p = wp->cr[p->n];
  who_q = wp->cr[q->n];

  if (who_p == NULL)
    return 1;
  if (who_q == NULL)
    return -1;

  if (who_p->wait < who_q->wait)
    return -1;
  if (who_p->wait > who_q->wait)
    return 1;

  return 0;
}

int
add_creature_here(world *wp, creature *who, int n)
{
  list_int *order = NULL;

  if (wp == NULL)
  {
    fprintf(stderr, "add_creature_here: wp is NULL\n");
    return -1;
  }
  if (who == NULL)
  {
    fprintf(stderr, "add_creature_here: who is NULL\n");
    return -1;
  }
  if (n < 0)
  {
    fprintf(stderr, "add_creature_here: strange n (%d)\n", n);
    return -1;
  }
  if (n == wp->player_id)
  {
    fprintf(stderr, "add_creature_here: n (%d) is for player\n", n);
    return -1;
  }

  while (n >= wp->cr_size)
  {
    if (world_expand_cr(wp) != 0)
    {  
      fprintf(stderr, "add_creature_here: world_expand_cr failed\n");
      return -1;
    }
  }

  if (wp->cr[n] != NULL)
  {
    fprintf(stderr, "add_creature_here: cr[%d] is already used\n", n);
    return -1;
  }

  order = list_int_new(n);
  if (order == NULL)
  {
    fprintf(stderr, "add_creature_here: list_int_new failed\n");
    return -1;
  }

  who->id = n;
  wp->cr[n] = who;
  wp->turn_order = list_int_merge(wp->turn_order, order, (void *) wp,
                                  &compare_wait);
  return n;
}

int
add_creature(world *wp, creature *who)
{
  int i;

  if (wp == NULL)
  {
    fprintf(stderr, "add_creature: wp is NULL\n");
    return -1;
  }
  if (who == NULL)
  {
    fprintf(stderr, "add_creature: who is NULL\n");
    return -1;
  }

  for (i = 0; i < wp->cr_size; i++)
  {
    if ((i != wp->player_id) && (wp->cr[i] == NULL))
      break;
  }

  return add_creature_here(wp, who, i);
}

int
remove_creature(world *wp, int creature_id)
{
  creature *who = NULL;
  list_int *order = NULL;
  list_int *order_temp = NULL;

  if (wp == NULL)
  {
    fprintf(stderr, "remove_creature: wp is NULL\n");
    return 1;
  }
  if ((creature_id < 0) || (creature_id >= wp->cr_size))
  {
    fprintf(stderr, "remove_creature: strange creature_id (%d)\n",
            creature_id);
    return 1;
  }
  if (wp->cr[creature_id] == NULL)
  {
    fprintf(stderr, "remove_creature: wp->cr[%d] is NULL\n", creature_id);
    return 1;
  }

  who = wp->cr[creature_id];

  order = wp->turn_order;
  while (order != NULL)
  {
    order_temp = order->next;
    if (order->n == who->id)
    {
      wp->turn_order = list_int_first(list_int_unlink(order));
      list_int_delete(order);
      order = NULL;
    }
    order = order_temp;
  }

  wp->cr[creature_id] = NULL;
  who->id = -1;
  creature_delete(who);
  who = NULL;

  return 0;
}

int
creature_is_dead(world *wp, int creature_id)
{
  if (wp == NULL)
  {
    fprintf(stderr, "creature_is_dead: wp is NULL\n");
    return 1;
  }
  if ((creature_id < 0) || (creature_id >= wp->cr_size))
  {
    fprintf(stderr, "creature_is_dead: strange creature_id (%d)\n",
            creature_id);
    return 1;
  }

  if (wp->cr[creature_id] == NULL)
    return 1;
  if (wp->cr[creature_id]->hp <= 0)
    return 1;

  return 0;
}

int
add_object_here(world *wp, object *what, int n)
{
  if (wp == NULL)
  {
    fprintf(stderr, "add_object_here: wp is NULL\n");
    return -1;
  }
  if (what == NULL)
  {
    fprintf(stderr, "add_object_here: what is NULL\n");
    return -1;
  }
  if (n < 0)
  {
    fprintf(stderr, "add_object_here: strange n (%d)\n", n);
    return -1;
  }

  while (n >= wp->obj_size)
  {
    if (world_expand_obj(wp) != 0)
    {
      fprintf(stderr, "add_object_here: world_expand_obj failed\n");
      return -1;
    }
  }

  if (wp->obj[n] != NULL)
  {
    fprintf(stderr, "add_object_here: obj[%d] is already used\n", n);
    return -1;
  }

  what->id = n;
  wp->obj[n] = what;

  return n;
}

int
add_object(world *wp, object *what)
{
  int i;

  if (wp == NULL)
  {
    fprintf(stderr, "add_object: wp is NULL\n");
    return -1;
  }
  if (what == NULL)
  {
    fprintf(stderr, "add_object: what is NULL\n");
    return -1;
  }

  for (i = 0; i < wp->obj_size; i++)
  {
    if (wp->obj[i] == NULL)
      break;
  }

  return add_object_here(wp, what, i);
}

int
remove_object(world *wp, int object_id)
{
  object *what = NULL;

  if (wp == NULL)
  {
    fprintf(stderr, "remove_object: wp is NULL\n");
    return 1;
  }
  if ((object_id < 0) || (object_id >= wp->obj_size))
  {
    fprintf(stderr, "remove_object: strange object_id (%d)\n", object_id);
    return 1;
  }
  if (wp->obj[object_id] == NULL)
  {
    fprintf(stderr, "remove_object: wp->obj[%d] is NULL\n", object_id);
    return 1;
  }

  what = wp->obj[object_id];

  wp->obj[object_id] = NULL;
  what->id = -1;
  object_delete(what);
  what = NULL;

  return 0;
}

int
object_is_dead(world *wp, int object_id)
{
  if (wp == NULL)
  {
    fprintf(stderr, "object_is_dead: wp is NULL\n");
    return 1;
  }
  if ((object_id < 0) || (object_id >= wp->obj_size))
  {
    fprintf(stderr, "object_is_dead: strange object_id (%d)\n",
            object_id);
    return 1;
  }

  if (wp->obj[object_id] == NULL)
    return 1;
  if (wp->obj[object_id]->quantity <= 0)
    return 1;

  return 0;
}

int
inventory_is_dead(world *wp, int inventory_id)
{
  if (wp == NULL)
  {
    fprintf(stderr, "inventory_is_dead: wp is NULL\n");
    return 1;
  }
  if ((inventory_id < 0) || (inventory_id >= 9))
  {
    fprintf(stderr, "inventory_is_dead: strange inventory_id (%d)\n",
            inventory_id);
    return 1;
  }

  if (wp->inv[inventory_id][0] < 0)
    return 1;
  if (wp->inv[inventory_id][1] <= 0)
    return 1;

  return 0;
}

int
grid_blocks_creature(world *wp, int x, int y)
{
  if (wp == NULL)
  {
    fprintf(stderr, "grid_block_creature: wp is NULL\n");
    return 1;
  }

  if ((x < 0) || (x >= wp->size_x))
    return 1;
  if ((y < 0) || (y >= wp->size_y))
    return 1;

  switch (wp->grid[x][y])
  {
  case 1:
    return 1;
    break;
  default:
    break;
  }

  return 0;
}

int
grid_has_creature(world *wp, int x, int y)
{
  int i;

  if (wp == NULL)
  {
    fprintf(stderr, "grid_has_creature: wp is NULL\n");
    return 1;
  }

  if ((x < 0) || (x >= wp->size_x))
    return 1;
  if ((y < 0) || (y >= wp->size_y))
    return 1;

  for (i = 0; i < wp->cr_size; i++)
  {
    if (wp->cr[i] == NULL)
      continue;
    if ((wp->cr[i]->x == x) && (wp->cr[i]->y == y))
    {
      return 1;
    }
  }

  return 0;
}

int
grid_blocks_los(world *wp, int x, int y)
{
  if (wp == NULL)
  {
    fprintf(stderr, "grid_blocks_los: wp is NULL\n");
    return 1;
  }

  if ((x < 0) || (x >= wp->size_x))
    return 1;
  if ((y < 0) || (y >= wp->size_y))
    return 1;

  if (wp->grid[x][y] == 1)
    return 1;

  return 0;
}

int
add_log(world *wp, const char *str)
{
  char *s;

  if (wp == NULL)
  {
    fprintf(stderr, "add_log: wp is NULL\n");
    return 1;
  }

  if (str == NULL)
    return 0;

  s = concat_string(1, str);
  if (s == NULL)
  {
    fprintf(stderr, "add_log: concat_string failed\n");
    return 1;
  }

  if ((wp->log_head < 0) || (wp->log_head >= wp->log_size))
  {
    fprintf(stderr, "add_log: strange log_head (%d)\n", wp->log_head);
    free(s);
    s = NULL;
    return 1;
  }
  if (wp->log[wp->log_head] != NULL)
  {
    free(wp->log[wp->log_head]);
    wp->log[wp->log_head] = NULL;
  }
  wp->log[wp->log_head] = s;
  wp->log_head = (wp->log_head + 1) % wp->log_size;

  return 0;
}
