/* $Id: loop.c,v 1.99 2010/03/05 01:46:41 oohara Exp $ */

#include <stdio.h>
#include <curses.h>
/* tigetnum */
#include <term.h>

#include "world.h"
#include "dungeon.h"
#include "creature.h"
#include "action.h"
#include "monster.h"
#include "los.h"

#include "loop.h"

static creature *find_creature_under_cursor(world *wp);
static int loop_input_0(world *wp);
static int loop_input_1(world *wp);
static int loop_input(world *wp);
static int loop_remove_dead(world *wp);
static int loop_action(world *wp);
static int loop_draw_0(world *wp);
static int loop_draw_1(world *wp);
static int loop_draw(world *wp);

static
creature *find_creature_under_cursor(world *wp)
{
  int i;
  creature *who = NULL;

  if (wp == NULL)
    return NULL;

  who = NULL;
  for (i = wp->cr_size - 1; i >= 0; i--)
  {
    if (creature_is_dead(wp, i))
      continue;
    if ((wp->cr[i]->x == wp->cursor_x)
        && (wp->cr[i]->y == wp->cursor_y))
      who = wp->cr[i];
  }

  return who;
}

static int
loop_input_0(world *wp)
{
  int c;
  int i;
  int x;
  int y;
  int n;
  int n0;
  creature *player = NULL;
  creature *who = NULL;

  if (wp == NULL)
    return 1;

  c = getch();

  switch (c)
  {
  case 'Q':
    return 1;
    break;
  case ' ':
    who = find_creature_under_cursor(wp);
    if (who == NULL)
      n0 = wp->player_id;
    else
      n0 = who->id;

    for (i = 1; i < wp->cr_size; i++)
    {
      n = (n0 + i) % wp->cr_size;
      if (n == wp->player_id)
        continue;
      if (creature_is_dead(wp, n))
        continue;
      x = wp->cr[n]->x;
      y = wp->cr[n]->y;
      if (distance(x, y, wp->camera_x, wp->camera_y) > 8)
        continue;
      wp->cursor_x = x;
      wp->cursor_y = y;
      break;
    }
    break;
  case 'i':
    wp->mode[0] = 1;
    wp->mode[1] = -1;
    break;
  default:
    break;
  }

  if (creature_is_dead(wp, wp->player_id))
    return 0;
  player = wp->cr[wp->player_id];

  switch (c)
  {
  case '7':
  case 'y':
    player->act[0] = ACTION_WALK;
    player->act[1] = -1;
    player->act[2] = -1;
    break;
  case '8':
  case 'k':
  case KEY_UP:
    player->act[0] = ACTION_WALK;
    player->act[1] = 0;
    player->act[2] = -1;
    break;
  case '9':
  case 'u':
    player->act[0] = ACTION_WALK;
    player->act[1] = 1;
    player->act[2] = -1;
    break;
  case '4':
  case 'h':
  case KEY_LEFT:
    player->act[0] = ACTION_WALK;
    player->act[1] = -1;
    player->act[2] = 0;
    break;
  case '5':
  case '.':
    player->act[0] = ACTION_STAY;
    break;
  case '6':
  case 'l':
  case KEY_RIGHT:
    player->act[0] = ACTION_WALK;
    player->act[1] = 1;
    player->act[2] = 0;
    break;
  case '1':
  case 'b':
    player->act[0] = ACTION_WALK;
    player->act[1] = -1;
    player->act[2] = 1;
    break;
  case '2':
  case 'j':
  case KEY_DOWN:
    player->act[0] = ACTION_WALK;
    player->act[1] = 0;
    player->act[2] = 1;
    break;
  case '3':
  case 'n':
    player->act[0] = ACTION_WALK;
    player->act[1] = 1;
    player->act[2] = 1;
    break;
  case 'Q':
    return 1;
    break;
  case 'S':
    player->act[0] = ACTION_SAVE;
    break;
  default:
    break;
  }

  if (!action_check(wp, player->id, player->act))
  {
    for (i = 0; i < 3; i++)
      player->act[i] = 0;
  }

  return 0;
}

static int
loop_input_1(world *wp)
{
  int c;
  creature *player = NULL;

  if (wp == NULL)
    return 1;

  c = getch();

  switch (c)
  {
  case 'i':
  case 'c':
    wp->mode[0] = 0;
    break;
  default:
    break;
  }

  if (creature_is_dead(wp, wp->player_id))
    return 0;
  player = wp->cr[wp->player_id];

  if (wp->mode[1] < 0)
  {
    switch (c)
    {
    case '7':
    case 'y':
      wp->mode[1] = 6;
      break;
    case '8':
    case 'k':
    case KEY_UP:
      wp->mode[1] = 7;
      break;
    case '9':
    case 'u':
      wp->mode[1] = 8;
      break;
    case '4':
    case 'h':
      wp->mode[1] = 3;
      break;
    case '5':
    case '.':
      wp->mode[1] = 4;
      break;
    case '6':
    case 'l':
    case KEY_RIGHT:
      wp->mode[1] = 5;
      break;
    case '1':
    case 'b':
      wp->mode[1] = 0;
      break;
    case '2':
    case 'j':
    case KEY_DOWN:
      wp->mode[1] = 1;
      break;
    case '3':
    case 'n':
      wp->mode[1] = 2;
      break;
    default:
      break;
    }
    if ((wp->mode[1] >= 0)
        && (inventory_is_dead(wp, wp->mode[1])))
      wp->mode[1] = -1;
    if ((wp->mode[1] >= 0)
        && (wp->inv[wp->mode[1]][0] == 0))
    {
      player->act[0] = ACTION_USE_OBJECT;
      player->act[1] = wp->mode[1];
      wp->mode[0] = 0;
    }
  }
  else
  {
    switch (c)
    {
    case '7':
    case 'y':
      player->act[0] = ACTION_USE_OBJECT;
      player->act[1] = wp->mode[1];
      player->act[2] = -1;
      player->act[3] = -1;
      wp->mode[0] = 0;
      break;
    case '8':
    case 'k':
    case KEY_UP:
      player->act[0] = ACTION_USE_OBJECT;
      player->act[1] = wp->mode[1];
      player->act[2] = 0;
      player->act[3] = -1;
      wp->mode[0] = 0;
      break;
    case '9':
    case 'u':
      player->act[0] = ACTION_USE_OBJECT;
      player->act[1] = wp->mode[1];
      player->act[2] = 1;
      player->act[3] = -1;
      wp->mode[0] = 0;
      break;
    case '4':
    case 'h':
      player->act[0] = ACTION_USE_OBJECT;
      player->act[1] = wp->mode[1];
      player->act[2] = -1;
      player->act[3] = 0;
      wp->mode[0] = 0;
      break;
    case '5':
    case '.':
      player->act[0] = ACTION_USE_OBJECT;
      player->act[1] = wp->mode[1];
      player->act[2] = 0;
      player->act[3] = 0;
      wp->mode[0] = 0;
      break;
    case '6':
    case 'l':
    case KEY_RIGHT:
      player->act[0] = ACTION_USE_OBJECT;
      player->act[1] = wp->mode[1];
      player->act[2] = 1;
      player->act[3] = 0;
      wp->mode[0] = 0;
      break;
    case '1':
    case 'b':
      player->act[0] = ACTION_USE_OBJECT;
      player->act[1] = wp->mode[1];
      player->act[2] = -1;
      player->act[3] = 1;
      wp->mode[0] = 0;
      break;
    case '2':
    case 'j':
    case KEY_DOWN:
      player->act[0] = ACTION_USE_OBJECT;
      player->act[1] = wp->mode[1];
      player->act[2] = 0;
      player->act[3] = 1;
      wp->mode[0] = 0;
      break;
    case '3':
    case 'n':
      player->act[0] = ACTION_USE_OBJECT;
      player->act[1] = wp->mode[1];
      player->act[2] = 1;
      player->act[3] = 1;
      wp->mode[0] = 0;
      break;
    default:
      break;
    }
  }

  return 0;
}

static int
loop_input(world *wp)
{
  if (wp == NULL)
  {
    fprintf(stderr, "loop_input: wp is NULL\n");
    return 1;
  }

  switch (wp->mode[0])
  {
  case 0:
    return loop_input_0(wp);
    break;
  case 1:
    return loop_input_1(wp);
    break;
  default:
    return 1;
    break;
  }

  return 1;
}

static int
loop_remove_dead(world *wp)
{
  int i;

  if (wp == NULL)
    return 1;

  for (i = 0; i < wp->cr_size; i++)
  {
    if (i == wp->player_id)
      continue;
    if (wp->cr[i] == NULL)
      continue;
    if (creature_is_dead(wp, i))
      remove_creature(wp, i);
  }

  for (i = 0; i < wp->obj_size; i++)
  {
    if (wp->obj[i] == NULL)
      continue;
    if (object_is_dead(wp, i))
      remove_object(wp, i);
  }

  for (i = 0; i < 8; i++)
  {
    if (!inventory_is_dead(wp, i))
      continue;
    wp->inv[i][0] = -1;
    wp->inv[i][1] = 0;
  }

  return 0;
}

static int
loop_action(world *wp)
{
  int i;
  int ut;
  creature *who = NULL;
  list_int *order = NULL;

  if (wp == NULL)
    return 1;

  do
  {
    loop_remove_dead(wp);

    if (creature_is_dead(wp, wp->player_id))
      return 0;

    if (wp->turn_order == NULL)
      return 0;
    who = wp->cr[wp->turn_order->n];
    if (who == NULL)
    {
      fprintf(stderr, "loop_action: who is NULL (%d)\n", wp->turn_order->n);
      return 1;
    }

    ut = who->wait;
    for (i = 0; i < wp->cr_size; i++)
    {
      if (wp->cr[i] == NULL)
        continue;
      wp->cr[i]->wait -= ut;
    }

    if (who->id != wp->player_id)
    {
      if (action_decide_monster(wp, who->id) != 0)
      {
        fprintf(stderr, "loop_action: action_decide_monster failed\n");
        return 1;
      }
    }

    ut = action_exec(wp, who->id);
    if (ut > 0)
    {
      order = wp->turn_order;
      wp->turn_order = list_int_unlink(order);
      who->wait += ut;
      wp->turn_order = list_int_merge(wp->turn_order, order, (void *) wp,
                                      &compare_wait);
    }
  } while (wp->turn_order->n != wp->player_id);

  loop_remove_dead(wp);

  return 0;
}

static int
loop_draw_0(world *wp)
{
  int i;
  int j;
  int x;
  int y;
  int attr;
  int scr_size_y;
  char buf[128];
  const char *s = NULL;
  creature *who = NULL;

  if (wp == NULL)
    return 1;

  scr_size_y = tigetnum("lines");

  if (distance(wp->cursor_x, wp->cursor_y,
               wp->camera_x, wp->camera_y) > 8)
  {
    if (!creature_is_dead(wp, wp->player_id))
    {
      wp->cursor_x = wp->cr[wp->player_id]->x;
      wp->cursor_y = wp->cr[wp->player_id]->y;
    }
  }
  if (distance(wp->cursor_x, wp->cursor_y,
               wp->camera_x, wp->camera_y) > 8)
  { 
    wp->cursor_x = wp->camera_x;
    wp->cursor_y = wp->camera_y;
  }

  clear();

  for (i = -6; i <= 6; i++)
  {
    for (j = -6; j <= 6; j++)
    {
      x = wp->camera_x + i;
      y = wp->camera_y + j;
      if ((x < 0) || (x >= wp->size_x))
        continue;
      if ((y < 0) || (y >= wp->size_y))
        continue;
      if (!has_los(wp, wp->camera_x, wp->camera_y, x, y))
        continue;
      switch (wp->grid[x][y])
      {
      case 0:
        attr = COLOR_PAIR(7);
        s = ".";
        break;
      case 1:
        attr = COLOR_PAIR(3);
        s = "#";
        break;
      default:
        attr = COLOR_PAIR(5) | A_BOLD;
        s = "~";
        break;
      }
      attron(attr);
      mvaddstr(j + 8, i + 8, s);
      attroff(attr);
    }
  }

  for (i = 0; i < wp->obj_size; i++)
  {
    if (object_is_dead(wp, i))
      continue;
    if (distance(wp->camera_x, wp->camera_y,
                 wp->obj[i]->x, wp->obj[i]->y) > 6)
      continue;
    if (!has_los(wp, wp->camera_x, wp->camera_y,
                 wp->obj[i]->x, wp->obj[i]->y))
      continue;
    switch (wp->obj[i]->type)
    {
    case 0:
      attr = COLOR_PAIR(7);
      s = "!";
      break;
    case 1:
      attr = COLOR_PAIR(7);
      s = "/";
      break;
    default:
      attr = COLOR_PAIR(5) | A_BOLD;
      s = "=";
      break;
    }
    attron(attr);
    mvaddstr(wp->obj[i]->y - wp->camera_y + 8,
             wp->obj[i]->x - wp->camera_x + 8,
             s);
    attroff(attr);
  }

  for (i = 0; i < wp->cr_size; i++)
  {
    if (creature_is_dead(wp, i))
      continue;
    if (distance(wp->cr[i]->x, wp->cr[i]->y,
                 wp->camera_x, wp->camera_y) > 8)
      continue;
    switch (wp->cr[i]->type)
    {
    case 0:
      attr = COLOR_PAIR(7) | A_BOLD;
      s = "@";
      break;
    case 1:
      attr = COLOR_PAIR(7);
      s = "g";
      break;
    default:
      attr = COLOR_PAIR(5) | A_BOLD;
      s = "^";
      break;
    }
    if ((distance(wp->camera_x, wp->camera_y,
                  wp->cr[i]->x, wp->cr[i]->y) > 6)
        || (!has_los(wp, wp->camera_x, wp->camera_y,
                     wp->cr[i]->x, wp->cr[i]->y)))
    {
      attr = COLOR_PAIR(8) | A_BOLD;
      s = "1";
      if ((wp->cr[i]->x == wp->cursor_x)
          && (wp->cr[i]->y == wp->cursor_y))
      {
        attr = COLOR_PAIR(7);
      }
    }
    attron(attr);
    mvaddstr(wp->cr[i]->y - wp->camera_y + 8,
             wp->cr[i]->x - wp->camera_x + 8,
             s);
    attroff(attr);
  }

  if (wp->cr[wp->player_id] != NULL)
  {
    attr = COLOR_PAIR(7);
    attron(attr);
    mvaddstr(0, 20, "player");
    sprintf(buf, "HP: %3d", wp->cr[wp->player_id]->hp);
    mvaddstr(1, 20, buf);
    attroff(attr);
  }

  who = find_creature_under_cursor(wp);
  if (who != NULL)
  {
    attr = COLOR_PAIR(7);
    attron(attr);
    if ((distance(wp->camera_x, wp->camera_y,
                  who->x, who->y) > 6)
        || (!has_los(wp, wp->camera_x, wp->camera_y,
                     who->x, who->y)))
    {
      mvaddstr(3, 20, "something");
    }
    else
    { 
      sprintf(buf, "creature #%d", who->id);
      mvaddstr(3, 20, buf);
      sprintf(buf, "HP: %3d", who->hp);
      mvaddstr(4, 20, buf);
      attroff(attr);
    }
  }

  switch (wp->mode[0])
  {
  case 0:  
    attr = COLOR_PAIR(7);
    attron(attr);
    for (i = 0;
         ((i < wp->log_size) && (scr_size_y - i - 2 >= 18));
         i++)
    {
      y = scr_size_y - i - 2;
      j = (wp->log_head - 1 - i + wp->log_size * 2) % wp->log_size;
      if (wp->log[j] == NULL)
        continue;
      mvaddstr(y, 0, wp->log[j]);
    }
    attroff(attr);
    break;
  case 1:
    attr = COLOR_PAIR(7);
    attron(attr);
    mvaddstr(18, 0, "use in which direction?");
    attroff(attr);
    break;
  default:
    break;
  }
  
  move(wp->cursor_y - wp->camera_y + 8,
       wp->cursor_x - wp->camera_x + 8);

  refresh();

  return 0;
}

static int
loop_draw_1(world *wp)
{
  int i;
  char buf[128];

  if (wp == NULL)
    return 1;

  clear();

  for (i = 0; i < 9; i++)
  {
    switch (i)
    {
    case 0:
      mvaddstr(i + 1, 0, "[b/1]");
      break;
    case 1:
      mvaddstr(i + 1, 0, "[j/2]");
      break;
    case 2:
      mvaddstr(i + 1, 0, "[n/3]");
      break;
    case 3:
      mvaddstr(i + 1, 0, "[h/4]");
      break;
    case 4:
      mvaddstr(i + 1, 0, "[./5]");
      break;
    case 5:
      mvaddstr(i + 1, 0, "[n/6]");
      break;
    case 6:
      mvaddstr(i + 1, 0, "[y/7]");
      break;
    case 7:
      mvaddstr(i + 1, 0, "[k/8]");
      break;
    case 8:
      mvaddstr(i + 1, 0, "[u/9]");
      break;
    default:
      break;
    }

    if ((wp->inv[i][0] < 0) || (wp->inv[i][1] <= 0))
      continue;

    sprintf(buf, "%2d", wp->inv[i][1]);
    mvaddstr(i + 1, 6, buf);

    switch (wp->inv[i][0])
    {
    case 0:
      mvaddstr(i + 1, 10, "potion");
      break;
    case 1:
      mvaddstr(i + 1, 10, "wand");
      break;
    default:
      mvaddstr(i + 1, 10, "junk");
      break;
    }
  }

  refresh();

  return 0;
}

static int
loop_draw(world *wp)
{
  if (wp == NULL)
    return 1;

  switch (wp->mode[0])
  {
  case 0:
    return loop_draw_0(wp);
    break;
  case 1:
    if (wp->mode[1] < 0)
      return loop_draw_1(wp);
    else
      return loop_draw_0(wp);
    break;
  default:
    return 1;
    break;
  }

  return 1;
}

int
main_loop(world *wp)
{
  if (wp == NULL)
  {
    fprintf(stderr, "main_loop: wp is NULL\n");
  }

  loop_draw(wp);

  while(1 == 1)
  {
    if (loop_input(wp) != 0)
      break;

    if (loop_action(wp) != 0)
      break;

    if (loop_draw(wp) != 0)
      break;
  }

  return 0;
}
