/* $Id: save.c,v 1.15 2010/03/05 08:56:03 oohara Exp $ */

/* FILE, rename */
#include <stdio.h>
/* stat */
#include <sys/stat.h>
/* errno */
#include <errno.h>

#include "world.h"
#include "int-loader.h"
#include "creature.h"
#include "object.h"
#include "list-int.h"
#include "dungeon.h"

#include "save.h"

#define SAVE_FILE_NAME "htwrl-save.txt"
#define OLD_SAVE_FILE_NAME "htwrl-save-old.txt"
#define TERMINATE_MARKER 0xdead

/* return 0 on success, 1 on error */
int
save_game(world *wp)
{
  int i;
  int j;
  int error_found;
  FILE *fp = NULL;
  creature *who = NULL;
  list_int *order = NULL;
  object *what = NULL;

  if (wp == NULL)
    return 1;

  fp = fopen(SAVE_FILE_NAME, "w");
  if (fp == NULL)
    return 1;

  error_found = 0;

  if (fprintf(fp, "# htwrl save file\n# DO NOT EDIT\n\n") < 0)
    error_found = 1;

  if (fprintf(fp, "# version\n0\n") < 0)
    error_found = 1;

  if (fprintf(fp, "\n# creature\n") < 0)
    error_found = 1;
  if (fprintf(fp, "## size\n%d\n", wp->cr_size) < 0)
    error_found = 1;
  for (i = 0; i < wp->cr_size; i++)
  {
    who = wp->cr[i];
    if (fprintf(fp, "# %d\n", i) < 0)
      error_found = 1;
    if (who == NULL)
    {
      if (fprintf(fp, "-1\n") < 0)
        error_found = 1;
      continue;
    }
    if (fprintf(fp, "%d\n%d\n%d\n%d\n%d\n%d\n",
                who->id, who->type, who->x, who->y,
                who->wait, who->hp) < 0)
      error_found = 1;
    for (j = 0; j < 4; j++)
    {
      if (fprintf(fp, "%d\n", who->act[j]) < 0)
        error_found = 1;
    }
  }
  if (fprintf(fp, "# creature end\n%#x\n", TERMINATE_MARKER) < 0)
    error_found = 1;

  if (fprintf(fp, "\n# turn_order\n") < 0)
    error_found = 1;
  if (fprintf(fp, "%d\n\n", list_int_length_all(wp->turn_order)) < 0)
    error_found = 1;
  order = wp->turn_order;
  while (order != NULL)
  {
    if (fprintf(fp, "%d\n", order->n) < 0)
      error_found = 1;
    order = order->next;
  }
  if (fprintf(fp, "# turn_order end\n%#x\n", TERMINATE_MARKER) < 0)
    error_found = 1;

  if (fprintf(fp, "\n# object\n") < 0)
    error_found = 1;
  if (fprintf(fp, "## size\n%d\n", wp->obj_size) < 0)
    error_found = 1;
  for (i = 0; i < wp->obj_size; i++)
  {
    what = wp->obj[i];
    if (fprintf(fp, "# %d\n", i) < 0)
      error_found = 1;
    if (what == NULL)
    {
      if (fprintf(fp, "-1\n") < 0)
        error_found = 1;
      continue;
    }
    if (fprintf(fp, "%d\n%d\n%d\n%d\n%d\n",
                what->id, what->type, what->x, what->y,
                what->quantity) < 0)
      error_found = 1;
  }
  if (fprintf(fp, "# object end\n%#x\n", TERMINATE_MARKER) < 0)
    error_found = 1;

  if (fprintf(fp, "\n# grid\n") < 0)
    error_found = 1;
  if (fprintf(fp, "## size\n%d\n%d\n", wp->size_x, wp->size_y) < 0)
    error_found = 1;
  for (i = 0; i < wp->size_x; i++)
  {
    if (fprintf(fp, "# (%d, y)\n", i) < 0)
      error_found = 1;
    for (j = 0; j < wp->size_y; j++)
    {
      if (fprintf(fp, "%d\n", wp->grid[i][j]) < 0)
        error_found = 1;
    }
  }
  if (fprintf(fp, "# grid end\n%#x\n\n", TERMINATE_MARKER) < 0)
    error_found = 1;

  if (fprintf(fp, "# player_id\n%d\n", wp->player_id) < 0)
    error_found = 1;
  if (fprintf(fp, "# camera_x\n%d\n", wp->camera_x) < 0)
    error_found = 1;
  if (fprintf(fp, "# camera_y\n%d\n", wp->camera_y) < 0)
    error_found = 1;

  if (fprintf(fp, "\n# inventory\n") < 0)
    error_found = 1;
  for (i = 0; i < 9; i++)
  {
    if (fprintf(fp, "# %d\n", i) < 0)
      error_found = 1;
    for (j = 0; j < 2; j++)
    {
      if (fprintf(fp, "%d\n", wp->inv[i][j]) < 0)
        error_found = 1;
    }
  }

  if (fprintf(fp, "\n# save file end\n%#x\n", TERMINATE_MARKER) < 0)
    error_found = 1;

  if (fclose(fp) != 0)
    error_found = 1;
  fp = NULL;

  if (error_found != 0)
    return 1;

  return 0;
}

/* return
 * 0 if a saved game is successfully loaded
 * 1 if no saved game is found
 * 2 on error
 */
int
load_game(world *wp)
{
  int i;
  int j;
  int m;
  int n;
  int error_found;
  struct stat buf_stat;
  int_loader *ilp = NULL;
  creature *who = NULL;
  object *what = NULL;

  if (wp == NULL)
    return 2;

  errno = 0;
  if (stat(SAVE_FILE_NAME, &buf_stat) != 0)
  {
    if (errno == ENOENT)
      return 1;
    return 2;
  }

  ilp = int_loader_new(SAVE_FILE_NAME);
  if (ilp == NULL)
    return 2;

  /* version */
  if (int_loader_get(ilp) != 0)
  {
    ilp->error_found = 1;
  }

  /* creature */
  m = int_loader_get(ilp);
  for (i = 0; i < m; i++)
  {
    n = int_loader_get(ilp);
    if (n < 0)
      continue;
    if (n != i)
    {
      ilp->error_found = 1;
      break;
    }
    who = creature_new();
    if (who == NULL)
    {
      ilp->error_found = 1;
      break;
    }
    who->id = n;
    who->type = int_loader_get(ilp);
    who->x = int_loader_get(ilp);
    who->y = int_loader_get(ilp);
    who->wait = int_loader_get(ilp);
    who->hp = int_loader_get(ilp);
    for (j = 0; j < 4; j++)
      who->act[j] = int_loader_get(ilp);
    if (add_creature_here(wp, who, n) < 0)
    {
      ilp->error_found = 1;
      creature_delete(who);
      who = NULL;
      break;
    }
  }
  if (int_loader_get(ilp) != TERMINATE_MARKER)
  {
      ilp->error_found = 1;
  }

  /* turn order */
  /* remove the list automatically created by add_creature_here() */
  if (wp->turn_order != NULL)
  {
    list_int_delete_all(wp->turn_order);
    wp->turn_order = NULL;
  }
  m = int_loader_get(ilp);
  for (i = 0; i < m; i++)
    wp->turn_order = list_int_append(wp->turn_order, int_loader_get(ilp));
  if (int_loader_get(ilp) != TERMINATE_MARKER)
  {
      ilp->error_found = 1;
  }

  /* object */
  m = int_loader_get(ilp);
  for (i = 0; i < m; i++)
  {
    n = int_loader_get(ilp);
    if (n < 0)
      continue;
    if (n != i)
    {
      ilp->error_found = 1;
      break;
    }
    what = object_new();
    if (what == NULL)
    {
      ilp->error_found = 1;
      break;
    }
    what->id = n;
    what->type = int_loader_get(ilp);
    what->x = int_loader_get(ilp);
    what->y = int_loader_get(ilp);
    what->quantity = int_loader_get(ilp);
    if (add_object_here(wp, what, n) < 0)
    {
      ilp->error_found = 1;
      object_delete(what);
      what = NULL;
      break;
    }
  }
  if (int_loader_get(ilp) != TERMINATE_MARKER)
  {
      ilp->error_found = 1;
  }

  /* grid */
  if (int_loader_get(ilp) != wp->size_x)
  {
      ilp->error_found = 1;
  }
  if (int_loader_get(ilp) != wp->size_y)
  {
      ilp->error_found = 1;
  }
  for (i = 0; i < wp->size_x; i++)
  {
    for (j = 0; j < wp->size_y; j++)
    {
      wp->grid[i][j] = int_loader_get(ilp);
    }
  }
  if (int_loader_get(ilp) != TERMINATE_MARKER)
  {
      ilp->error_found = 1;
  }

  wp->player_id = int_loader_get(ilp);
  wp->camera_x = int_loader_get(ilp);
  wp->camera_y = int_loader_get(ilp);

  /* inventory */
  for (i = 0; i < 9; i++)
  {
    for (j = 0; j < 2; j++)
    {
      wp->inv[i][j] = int_loader_get(ilp);
    }
  }

  /* save file end */
  if (int_loader_get(ilp) != TERMINATE_MARKER)
  {
      ilp->error_found = 1;
  }

  error_found = ilp->error_found;
  int_loader_delete(ilp);
  ilp = NULL;

  rename(SAVE_FILE_NAME, OLD_SAVE_FILE_NAME);

  if (error_found != 0)
    return 2;

  return 0;
}
