/* $Id: world.c,v 1.38 2010/03/05 00:47:40 oohara Exp $ */

#include <stdio.h>
/* malloc, atexit */
#include <stdlib.h>

#include <curses.h>

#include "creature.h"
#include "list-int.h"
#include "matrix.h"
#include "object.h"

#include "world.h"

static void curses_cleanup(void);

static
void curses_cleanup(void)
{
  clear();
  refresh();
  nocbreak();
  echo();
  endwin();
}

world *
world_new(void)
{
  int i;
  int j;
  world *wp = NULL;

  wp = (world *) malloc(sizeof(world));
  if (wp == NULL)
  {
    fprintf(stderr, "world_new: malloc(wp) failed\n");
    return NULL;
  }

  wp->cr_size = 0;
  wp->cr = NULL;
  wp->turn_order = NULL;
  wp->obj_size = 0;
  wp->obj = NULL;
  wp->size_x = 0;
  wp->size_y = 0;
  wp->grid = NULL;
  wp->player_id = -1;
  wp->camera_x = -1;
  wp->camera_y = -1;
  wp->inv = NULL;
  wp->scr = NULL;
  wp->mode = NULL;
  wp->cursor_x = -1;
  wp->cursor_y = -1;
  wp->log_size = 0;
  wp->log = NULL;
  wp->log_head = 0;

  wp->cr_size = 4;
  wp->cr = (creature **) malloc(sizeof(creature *) * (wp->cr_size));
  if (wp->cr == NULL)
  {
    fprintf(stderr, "world_new: malloc(wp->cr) failed\n");
    world_delete(wp);
    wp = NULL;
    return NULL;
  }
  for (i = 0; i < wp->cr_size; i++)
    wp->cr[i] = NULL;

  wp->obj_size = 4;
  wp->obj = (object **) malloc(sizeof(object *) * wp->obj_size);
  if (wp->obj == NULL)
  {
    fprintf(stderr, "world_new: malloc(wp->obj) failed\n");
    world_delete(wp);
    wp = NULL;
    return NULL;
  }
  for (i = 0; i < wp->obj_size; i++)
    wp->obj[i] = NULL;

  wp->size_x = 50;
  wp->size_y = 50;
  wp->grid = matrix2_new(wp->size_x, wp->size_y);
  if (wp->grid == NULL)
  {
    fprintf(stderr, "world_new: matrix2_new(wp->grid) failed\n");
    world_delete(wp);
    wp = NULL;
    return NULL;
  }
  for (i = 0; i < wp->size_x; i++)
  {
    for (j = 0; j < wp->size_y; j++)
    { 
      wp->grid[i][j] = -1;
    }
  }

  wp->inv = matrix2_new(9, 2);
  if (wp->inv == NULL)
  {
    fprintf(stderr, "world_new: malloc(wp->inv) failed\n");
    world_delete(wp);
    wp = NULL;
    return NULL;
  }
  for (i = 0; i < 9; i++)
  {
    wp->inv[i][0] = -1;
    wp->inv[i][1] = 0;
  }

  wp->mode = (int *) malloc(sizeof(int) * 2);
  if (wp->mode == NULL)
  {
    fprintf(stderr, "world_new: malloc(wp->mode) failed\n");
    world_delete(wp);
    wp = NULL;
    return NULL;
  }
  for (i = 0; i < 2; i++)
    wp->mode[i] = 0;

  wp->log_size = 50;
  wp->log = (char **) malloc(sizeof(char *) * wp->log_size);
  if (wp->log == NULL)
  {
    fprintf(stderr, "world_new: malloc(wp->log) failed\n");
    world_delete(wp);
    wp = NULL;
    return NULL;
  }
  for (i = 0; i < wp->log_size; i++)
    wp->log[i] = NULL;

  wp->scr = initscr();
  if (wp->scr == NULL)
  {
    fprintf(stderr, "world_new: initscr failed\n");
    world_delete(wp);
    wp = NULL;
    return NULL;
  }
  if (atexit(&curses_cleanup) != 0)
  {
    fprintf(stderr, "world_new: atexit failed\n");
    curses_cleanup();
    world_delete(wp);
    wp = NULL;
    return NULL;
  }

  start_color();
  noecho();
  cbreak();
  keypad(wp->scr, TRUE);

  init_pair(1, COLOR_RED, COLOR_BLACK);
  init_pair(2, COLOR_GREEN, COLOR_BLACK);
  init_pair(3, COLOR_YELLOW, COLOR_BLACK);
  init_pair(4, COLOR_BLUE, COLOR_BLACK);
  init_pair(5, COLOR_MAGENTA, COLOR_BLACK);
  init_pair(6, COLOR_CYAN, COLOR_BLACK);
  init_pair(7, COLOR_WHITE, COLOR_BLACK);
  init_pair(8, COLOR_BLACK, COLOR_BLACK);

  return wp;
}

void
world_delete(world *wp)
{
  int i;

  if (wp == NULL)
    return;

  if (wp->cr != NULL)
  {
    for (i = 0; i < wp->cr_size; i++)
    {
      if (wp->cr[i] != NULL)
      {
        creature_delete(wp->cr[i]);
        wp->cr[i] = NULL;
      }
    }
    free(wp->cr);
    wp->cr = NULL;
    wp->cr_size = 0;
  }

  if (wp->turn_order != NULL)
  {
    list_int_delete_all(wp->turn_order);
    wp->turn_order = NULL;
  }

  if (wp->obj != NULL)
  {
    for (i = 0; i < wp->obj_size; i++)
    {
      if (wp->obj[i] != NULL)
      {
        object_delete(wp->obj[i]);
        wp->obj[i] = NULL;
      }
    }
    free(wp->obj);
    wp->obj = NULL;
    wp->obj_size = 0;
  }

  if (wp->grid != NULL)
  {
    matrix2_delete(wp->grid, wp->size_x, wp->size_y);
    free(wp->grid);
    wp->grid = NULL;
    wp->size_x = 0;
    wp->size_y = 0;
  }

  if (wp->inv != NULL)
  {
    matrix2_delete(wp->inv, 9, 2);
    free(wp->inv);
    wp->inv = NULL;
  }

  if (wp->mode != NULL)
  {
    free(wp->mode);
    wp->mode = NULL;
  }

  if (wp->log != NULL)
  {
    for (i = 0; i < wp->log_size; i++)
    {
      if (wp->log[i] != NULL)
      {
        free(wp->log[i]);
        wp->log[i] = NULL;
      }
    }
    free(wp->log);
    wp->log = NULL;
    wp->log_size = 0;
  }

  free(wp);
  wp = NULL;
}

void
world_reset_game(world *wp)
{
  int i;
  int j;

  if (wp == NULL)
  {
    fprintf(stderr, "world_reset_game: wp is NULL\n");
    return;
  }

  for (i = 0; i < wp->cr_size; i++)
  {
    if (wp->cr[i] != NULL)
    {
      creature_delete(wp->cr[i]);
      wp->cr[i] = NULL;
    }
  }

  if (wp->turn_order != NULL)
  {
    list_int_delete_all(wp->turn_order);
    wp->turn_order = NULL;
  }

  for (i = 0; i < wp->obj_size; i++)
  {
    if (wp->obj[i] != NULL)
    {
      object_delete(wp->obj[i]);
      wp->obj[i] = NULL;
    }
  }

  for (i = 0; i < wp->size_x; i++)
  {
    for (j = 0; j < wp->size_y; j++)
    {
      wp->grid[i][j] = -1;
    }
  }

  wp->player_id = -1;
  wp->camera_x = -1;
  wp->camera_y = -1;

  for (i = 0; i < 9; i++)
  {
    wp->inv[i][0] = -1;
    wp->inv[i][1] = 0;
  }

  for (i = 0; i < 2; i++)
    wp->mode[i] = 0;

  wp->cursor_x = -1;
  wp->cursor_y = -1;

  for (i = 0; i < wp->log_size; i++)
  {
    if (wp->log[i] != NULL)
    {
      free(wp->log[i]);
      wp->log[i] = NULL;
    }
  }
  wp->log_head = 0;
}

int
world_expand_cr(world *wp)
{
  int i;
  int cr_size_temp;
  creature **cr_temp = NULL;

  if (wp == NULL)
  {
    fprintf(stderr, "world_expand_cr: p is NULL\n");
    return 1;
  }

  cr_size_temp = wp->cr_size * 2;
  if (cr_size_temp <= 0)
  {
    fprintf(stderr, "world_expand_cr: cr_size overflowed\n");
    return 1;
  }
  cr_temp = (creature **) realloc(wp->cr, sizeof(creature *) * cr_size_temp);
  if (cr_temp == NULL)
  {
    fprintf(stderr, "world_expand_cr: realloc failed\n");
    return 1;
  }

  wp->cr = cr_temp;
  for (i = wp->cr_size; i < cr_size_temp; i++)
    wp->cr[i] = NULL;
  wp->cr_size = cr_size_temp;

  return 0;
}

int
world_expand_obj(world *wp)
{
  int i;
  int obj_size_temp;
  object **obj_temp = NULL;

  if (wp == NULL)
  {
    fprintf(stderr, "world_expand_obj: wp is NULL\n");
    return 1;
  }

  obj_size_temp = wp->obj_size * 2;
  if (obj_size_temp <= 0)
  {
    fprintf(stderr, "world_expand_obj: obj_size overflowed\n");
    return 1;
  }
  obj_temp = (object **) realloc(wp->obj, sizeof(object *) * obj_size_temp);
  if (obj_temp == NULL)
  {
    fprintf(stderr, "world_expand_obj: realloc failed\n");
    return 1;
  }

  wp->obj = obj_temp;
  for (i = wp->obj_size; i < obj_size_temp; i++)
    wp->obj[i] = NULL;
  wp->obj_size = obj_size_temp;

  return 0;
}
